from dgg_chat.exceptions import InvalidMessageError, AnonymousConnectionError
from ._exceptions import (
    CommandDescriptionTooLongError,
    DuplicateCommandError,
    InvalidCommandError,
    InvalidCommandArgumentsError,
    UnknownCommandError,
)
